/*

    File: photorec.c

    Copyright (C) 1998-2005 Christophe GRENIER <grenier@cgsecurity.org>
  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#include <stdarg.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>	/* geteuid */
#endif
#ifdef HAVE_STRING_H
#include <string.h>
#endif
#ifdef HAVE_TIME_H
#include <time.h>
#endif
#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif
#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif
#include <ctype.h>      /* toupper, tolower */
#ifdef HAVE_LOCALE_H
#include <locale.h>	/* setlocale */
#endif
#ifdef HAVE_SIGNAL
#include <signal.h>
#endif
#ifdef HAVE_SYS_WAIT_H
#include <sys/wait.h>
#endif
#ifdef HAVE_SETJMP_H
#include <setjmp.h>
#endif
#include <errno.h>
#include "types.h"
#ifdef HAVE_JPEGLIB_H
#include <jpeglib.h>
#endif
#include "common.h"
#include "lang.h"
#include "intrf.h"
#include "godmode.h"
#include "fnctdsk.h"

extern const t_arch_fnct arch_i386;
extern const t_arch_fnct arch_sun;
extern const t_arch_fnct arch_none;
#define INTER_SELECT_X	0
#define INTER_SELECT_Y	23
#define INTER_SELECT	15
enum state_type { STATE_NONE, STATE_OPEN, STATE_OPENJPG, STATE_OPENCRW, STATE_WRITE, STATE_WRITEJPG, STATE_WRITECRW };
typedef enum state_type state_type_t;
typedef struct struct_alloc_data t_alloc_data;
struct struct_alloc_data
{
  uint64_t start;
  uint64_t end;
  t_alloc_data *prev;
  t_alloc_data *next;
  const char *extension;
};

static void interface_select(t_param_disk *disk_car, const int debug,const int paranoid, const char *recup_dir, const int keep_corrupted_file);
static void interface_options_photorec(int *paranoid, int *allow_partial_last_cylinder,int *halt_on_errors, const t_arch_fnct **arch, int *keep_corrupted_file);
static int do_curses_photorec(int debug, const char *recup_dir, const t_list_disk *list_disk, const t_arch_fnct **arch);
static int photorec(t_param_disk *disk_car, t_partition *partition, const int debug, const int paranoid, const char *recup_dir, const int keep_corrupted_file);

static FILE* init_log(const char*filename,int argc, char**argv);

static FILE *f_rapport=NULL;
static int f_status=0;
static int64_t test_jpeg(FILE *infile);
static unsigned int find_blocksize(t_alloc_data *list_file, const unsigned int default_blocksize);
static void update_blocksize(unsigned int blocksize, t_alloc_data *list_free_space);
static t_alloc_data *update_free_space(const char *filename_out, const char*extension, t_alloc_data *list_free_space, t_alloc_data **new_current_free_space, const uint64_t old_start, uint64_t file_size, const unsigned int blocksize, const unsigned int sector_size,const unsigned int space_is_allocated);

void ecrit_rapport_string(const char *string,const int max_length)
{
  int i;
  for(i=0;(string[i]!='\0')&&(i<max_length);i++)
    ecrit_rapport("%c",string[i]);
}

int ecrit_rapport(const char *_format, ...)
{
  int res=0;
  if(f_rapport!=NULL)
  {
    va_list ap;
    fflush(stderr);
    va_start(ap,_format);
    res=vfprintf(f_rapport,_format,ap);
    va_end(ap);
    if(res<0)
    {
      f_status=1;
    }
    if(fflush(f_rapport))
    {
      f_status=1;
    }
  }
  return res;
}

static FILE* init_log(const char*filename,int argc, char**argv)
{
  FILE*f_file=fopen(filename,"a");
  if(f_file==NULL)
  {
    printf("\nCan not create photorec.log: %s\n",strerror(errno));
    exit(1);
  }
  else
  {
	int i;
	time_t my_time;
	my_time=time(NULL);
	fprintf(f_file,"\n\n%s",ctime(&my_time));
	fprintf(f_file,"PhotoRec command line :");
	for(i=1;i<argc;i++)
	  fprintf(f_file," %s", argv[i]);
	fprintf(f_file,"\n");
	fflush(f_file);
#ifdef HAVE_DUP2
	dup2(fileno(f_file),2);
#endif
  }
  return f_file;
}

#ifdef HAVE_SIGACTION
void sighup_hdlr(int shup)
{
  ecrit_rapport("SIGHUP detected! PhotoRec has been killed.\n");
  if(f_rapport!=NULL)
  {
    fclose(f_rapport);
  }
  exit(1);
}
#endif
static t_alloc_data *update_free_space(const char *filename_out,const char *extension, t_alloc_data *list_free_space, t_alloc_data **new_current_free_space, const uint64_t file_start, uint64_t file_size, const unsigned int blocksize, const unsigned int sector_size,const unsigned int space_is_allocated)
{
  t_alloc_data *current_free_space;
  file_size=(file_size+blocksize-1)/blocksize*blocksize;
/*  ecrit_rapport("update_free_space(filename_out=%s,file_start=%lu, file_size=%lu)",filename_out,(long unsigned)(file_start/sector_size),(long unsigned)(file_size/sector_size)); */
  for(current_free_space=list_free_space;current_free_space!=NULL && (current_free_space->next!=NULL && current_free_space->next->start<=file_start);current_free_space=current_free_space->next);
  if(current_free_space->start<file_start)
  {
    t_alloc_data *new_free_space;
    new_free_space=(t_alloc_data*)MALLOC(sizeof(*new_free_space));
    new_free_space->start=file_start;
    new_free_space->end=current_free_space->end;
    new_free_space->prev=current_free_space;
    new_free_space->next=current_free_space->next;
    new_free_space->extension=NULL;
    current_free_space->end=file_start-1;
    current_free_space->next=new_free_space;
    current_free_space=new_free_space;
  }
  if(current_free_space!=NULL && current_free_space->start==file_start)
  {
    current_free_space->extension=extension;
  }
  ecrit_rapport("%s",filename_out);
  while(current_free_space!=NULL && current_free_space->end-current_free_space->start+1<=file_size)
  {
    t_alloc_data *next;
    next=current_free_space->next;
    file_size-=current_free_space->end-current_free_space->start+1;
    ecrit_rapport(" %lu-%lu",(long unsigned)(current_free_space->start/sector_size),
	(long unsigned)(current_free_space->end/sector_size));
    if(space_is_allocated>0)
    {
      if(current_free_space->prev!=NULL)
	current_free_space->prev->next=current_free_space->next;
      if(current_free_space->next!=NULL)
	current_free_space->next->prev=current_free_space->prev;
      if(list_free_space==current_free_space)
	list_free_space=next;
      FREE(current_free_space);
    }
    current_free_space=next;
  }
  if(current_free_space!=NULL && file_size>0)
  {
    ecrit_rapport(" %lu-%lu",(long unsigned)(current_free_space->start/sector_size),
	(long unsigned)((current_free_space->start+file_size-1)/sector_size));
    if(space_is_allocated>0)
    {
      current_free_space->start+=file_size;
      current_free_space->extension=NULL;
    }
  }
  if(space_is_allocated>0)
  {
    *new_current_free_space=current_free_space;
    ecrit_rapport("\n");
  }
  else
  {
    ecrit_rapport(" failed\n");
  }
  return list_free_space;
}

static unsigned int menu_choose_blocksize(unsigned int blocksize, const unsigned int sector_size)
{
  /* TODO: let the user choose the blocksize */
  /* return sector_size; */
  return blocksize;
}

static int photorec(t_param_disk *disk_car, t_partition *partition, const int debug, const int paranoid, const char *recup_dir, const int keep_corrupted_file)
{
  time_t real_start_time;
  t_alloc_data *list_free_space;
  t_alloc_data *current_free_space;
  t_alloc_data *list_file=NULL;
  unsigned int file_nbr=0;
  char filename_out[2048];
  state_type_t state=STATE_NONE;
  FILE *file_out=NULL;
  int ind_stop=0;
  unsigned int pass;
  unsigned int blocksize=disk_car->sector_size;
  const unsigned char jpg_header[4]= { 0xff,0xd8,0xff,0xe1};
  const unsigned char jpg2_header[4]= { 0xff,0xd8,0xff,0xe0};
  const unsigned char mov_header[7]= { 0x00,0x00,0x00,0x14,0x70,0x6E,0x6f};	/* Bad definition */
  const unsigned char mpg_header[4]= { 0x00,0x00,0x01,0xbA};
  const unsigned char mrw_header[4]= { 0x00,'M','R','M'};			/* Minolta Raw */
#ifdef RECOVER_MP3
  const unsigned char mp3_header[2]= { 0xFF, 0xFB };
  /* mp3 recovery is not a very good idea because it has been designed for streaming so a single mp3 file will be recovered in a lot of files */
#endif

  filename_out[sizeof(filename_out)-1]='\0';
  aff_buffer(BUFFER_RESET,"Q");
  aff_copy(stdscr);
  wmove(stdscr,4,0);
  wdoprintf(stdscr,"%s",disk_car->description(disk_car));
  mvwaddstr(stdscr,5,0,msg_PART_HEADER_LONG);
  wmove(stdscr,6,0);
  aff_part(stdscr,AFF_PART_ORDER,disk_car,partition);
  wrefresh(stdscr);
  wmove(stdscr,22,0);
  wattrset(stdscr,A_STANDOUT);
  waddstr(stdscr,"  Stop  ");
  wattroff(stdscr,A_STANDOUT);
  ecrit_rapport("\nAnalyse ");
  aff_part_rapport(disk_car,partition);
  if(debug>0)
  {
    ecrit_rapport("debug=%u, paranoid=%u\n",debug,paranoid);
  }
  list_free_space=(t_alloc_data*)MALLOC(sizeof(*list_free_space));
  list_free_space->start=partition->part_offset;
  list_free_space->end=partition->part_offset+partition->part_size-1;
  list_free_space->extension=NULL;
  list_free_space->prev=NULL;
  list_free_space->next=NULL;
  real_start_time=time(NULL);
  for(pass=1;ind_stop==0 || ind_stop==2;pass++)
  {
    const char *extension=NULL;
    const char *old_extension=NULL;
    uint64_t offset;
    uint64_t start=0;
    uint64_t file_size=0;
    int old_state=STATE_NONE;
    int dont_keep_file=0;
    unsigned int old_file_nbr=file_nbr;
    unsigned char real_buffer[0x20+blocksize];
    unsigned char *buffer=&real_buffer[0x20];
    time_t start_time;
    time_t current_time;
    time_t previous_time;
    start_time=time(NULL);
    previous_time=start_time;
    memset(real_buffer,0,0x20+blocksize);
    ecrit_rapport("Pass %u (blocksize=%u)\n",pass,blocksize);
    if(debug>1)
    {
      t_alloc_data *cur_free_space;
      ecrit_rapport("Explore ");
      for(cur_free_space=list_free_space;cur_free_space!=NULL;cur_free_space=cur_free_space->next)
	ecrit_rapport(" %lu-%lu",(long unsigned)(cur_free_space->start/disk_car->sector_size),
	    (long unsigned)(cur_free_space->end/disk_car->sector_size));
      ecrit_rapport("\n");
    }
    {
      static uint64_t old_offset;
      uint64_t old_start=0;
      if(pass!=2)
	old_offset=0;
      else
	old_offset-=blocksize;
      for(current_free_space=list_free_space;current_free_space!=NULL && (ind_stop==0 || ind_stop==2);)
      {
	for(offset=current_free_space->start;current_free_space!=NULL && (offset+blocksize<current_free_space->end) && (ind_stop==0 || ind_stop==2) &&
	    ((pass==1 && file_nbr<10)||(pass==2)||(pass>2 && (state==STATE_WRITE || state==STATE_WRITEJPG || current_free_space->extension!=NULL)));
	    offset+=blocksize)
	{
	  if(offset>old_offset)
	  {
	    old_offset=offset;
	    if(debug>1)
	    {
	      ecrit_rapport("Reading sector %10lu/%lu\n",(unsigned long)((offset-partition->part_offset)/disk_car->sector_size),(unsigned long)(partition->part_size/disk_car->sector_size));
	    }
	    if(disk_car->read(disk_car,blocksize, buffer, offset)!=0)
	    { /* Read error */
	      memset(real_buffer,0,0x20+blocksize);
	    }
	    {
	      unsigned int size_of_data_to_write=blocksize;
	      old_state=state;
	      if(offset%(123*0x200)==0)
	      {
		current_time=time(NULL);
		if(offset!=partition->part_offset && current_time>previous_time)
		{
		  previous_time=current_time;
		  wmove(stdscr,9,0);
		  wclrtoeol(stdscr);
		  wdoprintf(stdscr,"Reading sector %10lu/%lu, %u pictures or videos found\n",(unsigned long)((offset-partition->part_offset)/disk_car->sector_size),(unsigned long)(partition->part_size/disk_car->sector_size),file_nbr);
		  wmove(stdscr,10,0);
		  wclrtoeol(stdscr);
		  wdoprintf(stdscr,"Elapsed time %uh%02um%02us - Estimated time for achievement %uh%02um%02u\n",
		      (unsigned)((current_time-real_start_time)/60/60),
		      (unsigned)((current_time-real_start_time)/60%60),
		      (unsigned)((current_time-real_start_time)%60),
		      (unsigned)((partition->part_offset+partition->part_size-1-offset)*(current_time-real_start_time)/(offset-partition->part_offset)/3600),
		      (unsigned)(((partition->part_offset+partition->part_size-1-offset)*(current_time-real_start_time)/(offset-partition->part_offset)/60)%60),
		      (unsigned)((partition->part_offset+partition->part_size-1-offset)*(current_time-real_start_time)/(offset-partition->part_offset))%60);
		  wrefresh(stdscr);
		  switch(wgetch_nodelay(stdscr))
		  {
		    case KEY_ENTER:
#ifdef PADENTER
		    case PADENTER:
#endif
		    case '\n':
		    case '\r':
		    case 's':
		    case 'S':
		      ind_stop=1;
		      break;
		  }
		}
	      }
	      if((memcmp(buffer,jpg_header,sizeof(jpg_header))==0)
		  ||(memcmp(buffer,jpg2_header,sizeof(jpg2_header))==0))
	      {
		old_start=start;
		old_extension=extension;
		start=offset;
		extension="jpg";
		state=STATE_OPENJPG;
	      }
	      else if(memcmp(buffer,mov_header,sizeof(mov_header))==0)
	      {
		old_start=start;
		old_extension=extension;
		start=offset;
		extension="mov";
		state=STATE_OPEN;
	      }
	      else if(memcmp(buffer,mpg_header,sizeof(mpg_header))==0)
	      {
		if((extension && strcmp(extension,"mpg"))|| state!=STATE_WRITE)
		{
		  old_start=start;
		  old_extension=extension;
		  start=offset;
		  extension="mpg";
		  state=STATE_OPEN;
		}
	      }
	      else if(memcmp(buffer,mrw_header,sizeof(mrw_header))==0)
	      {
		/* Minolta */
		old_start=start;
		old_extension=extension;
		start=offset;
		extension="mrw";
		state=STATE_OPEN;
	      }
	      else if(((buffer[0]==0x49 && buffer[1]==0x49)||(buffer[0]==0x4D && buffer[1]==0x4D))
		  && memcmp(buffer+6,"HEAPCCDR",8)==0)
	      {
		/* Canon */
		old_start=start;
		old_extension=extension;
		start=offset;
		extension="crw";
		state=STATE_OPENCRW;
	      }
	      else if(memcmp(buffer,"RIFF",4)==0)
	      {
		old_start=start;
		old_extension=extension;
		start=offset;
		extension="avi";	/* can also be some wav files */
		state=STATE_OPEN;
	      }
	      else if(memcmp(buffer,"FOVb",4)==0)
	      {
		/* Sigma/Foveon */
		old_start=start;
		old_extension=extension;
		start=offset;
		extension="x3f";
		state=STATE_OPEN;
	      }
	      else if(memcmp (buffer+25, "ARECOYK", 7)==0)
	      {
		/* Contax */
		old_start=start;
		old_extension=extension;
		start=offset;
		extension="raw";	/* What is the correct extension ? */
		state=STATE_OPEN;
	      }
	      else if(memcmp (buffer, "FUJIFILM", 8)==0)
	      {
		/* Fuji */
		old_start=start;
		old_extension=extension;
		start=offset;
		extension="raf";
		state=STATE_OPEN;
	      }
	      else if(memcmp (buffer, "DSC-Image", 9)==0)
	      {
		/* Rollei */
		old_start=start;
		old_extension=extension;
		start=offset;
		extension="rdc";
		state=STATE_OPEN;
	      }
#ifdef RECOVER_MP3
	      else if(memcmp(buffer,mp3_header,sizeof(mp3_header))==0)
	      {
		if((extension && strcmp(extension,"mp3"))|| state!=STATE_WRITE)
		{
		  old_start=start;
		  old_extension=extension;
		  extension="mp3";
		  state=STATE_OPEN;
		}
	      }
#endif
	      else if(memcmp(buffer,"C:\\DCIM\\",8)==0)
	      {
		old_start=start;
		old_extension=extension;
		start=offset;
		extension="ctg";
		state=STATE_OPEN;
		dont_keep_file=1;
	      }
	      else if(strncmp(&buffer[0],".          ",8+3)==0 && strncmp(&buffer[0x20],"..         ",8+3)==0)
	      {
		old_start=start;
		old_extension=extension;
		start=offset;
		extension="dir";
		state=STATE_OPEN;
		dont_keep_file=1;
	      }
	      else if(state==STATE_WRITECRW)
	      {
		unsigned int i;
		for(i=0;i+13<blocksize+0x20 && state==STATE_WRITECRW;i++)
		{
		  if(real_buffer[i]==0x0A && real_buffer[i+1]==0x30 && real_buffer[i+2]==0xFE && real_buffer[i+3]==0x17 && real_buffer[i+13]==0)
		  {
		    old_start=start;
		    old_extension=extension;
		    state=STATE_NONE;
		    size_of_data_to_write=i+14-0x20;
		    fwrite(buffer,size_of_data_to_write,1,file_out);
		    file_size+=size_of_data_to_write;
		    extension=NULL;
		  }
		}
	      }
	      else if(state==STATE_WRITEJPG && file_size>=10*1024*1024)
	      { /* The JPG file is too big */
		old_start=start;
		old_extension=extension;
		state=STATE_NONE;
		extension=NULL;
	      }
	      else if(state==STATE_WRITE && extension!=NULL)
	      {
		if( (memcmp(extension,"ctg",3)==0 && file_size>=4634) ||
		    (memcmp(extension,"dir",3)==0 && file_size>=blocksize))
		{
		  old_start=start;
		  old_extension=extension;
		  state=STATE_NONE;
		  extension=NULL;
		}
	      }
	      if(state==STATE_OPEN || state==STATE_OPENJPG || state==STATE_OPENCRW || state==STATE_NONE)
	      {
		if(file_out)
		{
		  if((old_state==STATE_WRITEJPG) && paranoid>0)
		  {
		    int64_t jpg_file_size=test_jpeg(file_out);
		    if(jpg_file_size<0)
		    {
		      update_free_space(filename_out,old_extension,list_free_space,&current_free_space,old_start,file_size,blocksize,disk_car->sector_size,0);
		    }
		    file_size=(jpg_file_size>0?jpg_file_size:0);
		  }
		  if(ind_stop!=2)
		  {
#ifdef HAVE_FTRUNCATE
		    ftruncate(fileno(file_out), file_size);
#endif
		    list_free_space=update_free_space(filename_out,old_extension,list_free_space,&current_free_space,old_start,file_size,blocksize,disk_car->sector_size,(file_size>0));
		  }
		  if(file_size>0 && pass==1)
		  {
		    /* Backup file offset */
		    t_alloc_data *new_file_alloc;
		    new_file_alloc=(t_alloc_data*)MALLOC(sizeof(*new_file_alloc));
		    new_file_alloc->start=old_start;
		    new_file_alloc->end=0;
		    new_file_alloc->prev=NULL;
		    new_file_alloc->next=list_file;
		    if(list_file!=NULL)
		      list_file->prev=new_file_alloc;
		    list_file=new_file_alloc;
		  }
		  if((file_size>0 || ind_stop==2) && dont_keep_file==0)
		  {
		    file_nbr++;
		  }
		  else
		  {
		    dont_keep_file=0;
		    unlink(filename_out);
		  }
		  fclose(file_out);
		  file_out=NULL;
		}
	      }
	      if(state==STATE_OPEN || state==STATE_OPENJPG || state==STATE_OPENCRW)
	      {
		snprintf(filename_out,sizeof(filename_out)-1,"%s/%d.%s",recup_dir,file_nbr+1,extension);
		file_size=0;
		if(debug>1)
		{
		  ecrit_rapport("header found %s, sector %lu\n",filename_out,
		      (unsigned long)((offset-partition->part_offset)/disk_car->sector_size));
		}
		if(!(file_out=fopen(filename_out,"w+b")))
		{
		  perror("out");
		  state=STATE_NONE;
		}
		else
		{
		  switch(state)
		  {
		    case STATE_OPENJPG: state=STATE_WRITEJPG; break;
		    case STATE_OPENCRW: state=STATE_WRITECRW; break;
		    default: state=STATE_WRITE; break;
		  }
		}
	      }
	      if(file_out)
	      {
		/*	    ecrit_rapport("%s add %lu bytes from sector %lu\n",filename_out, (unsigned long) size_of_data_to_write,
			    (unsigned long)((offset-partition->part_offset)/disk_car->sector_size));
		 */
		fwrite(buffer,size_of_data_to_write,1,file_out);
		file_size+=size_of_data_to_write;
	      }
	      memcpy(&real_buffer,buffer,0x20);
	    }
	  }
	}
	if(current_free_space!=NULL)
	  current_free_space=current_free_space->next;
      }
    } /* Pass is finished */
    if(file_out)
    {
      if(state==STATE_WRITEJPG && paranoid>0)
      {
	int64_t jpg_file_size=test_jpeg(file_out);
	if(jpg_file_size<0)
	{
	  update_free_space(filename_out,old_extension,list_free_space,&current_free_space,start,file_size,blocksize,disk_car->sector_size,0);
	}
	file_size=(jpg_file_size>0?jpg_file_size:0);
      }
#ifdef OLD
      else if(state==STATE_WRITECRW)
      { /* Already have the file_size */
      }
      else
	file_size=ftell(file_out);
#endif
      /* ecrit_rapport("file_size=%lu\n",(long unsigned)file_size); */
      if(ind_stop!=2)
      {
#ifdef HAVE_FTRUNCATE
	ftruncate(fileno(file_out), file_size);
#endif
	list_free_space=update_free_space(filename_out,extension,list_free_space,&current_free_space,start,file_size,blocksize,disk_car->sector_size,(file_size>0));
      }
      if((file_size>0 || ind_stop==2) && dont_keep_file==0)
      {
	file_nbr++;
      }
      else
      {
	dont_keep_file=0;
	unlink(filename_out);
      }
      fclose(file_out);
      file_out=NULL;
      file_size=0;
    }
    if(ind_stop==2)
    {
      ind_stop=3;
    }
    else if(old_file_nbr==file_nbr || (paranoid==0 && pass==2))
    {
      if(keep_corrupted_file>0)
	ind_stop=2;
      else
	ind_stop=3;
    }
    else
    {
      if(pass==1)
      {
	blocksize=find_blocksize(list_file,disk_car->sector_size);
	if(blocksize>disk_car->sector_size)
	{
	  blocksize=menu_choose_blocksize(blocksize, disk_car->sector_size);
	}
	update_blocksize(blocksize,list_free_space);
      }
    }
    ecrit_rapport("Elapsed time %uh%02um%02us\n",
	(unsigned)((current_time-real_start_time)/60/60),
	(unsigned)((current_time-real_start_time)/60%60),
	(unsigned)((current_time-real_start_time)%60));
    ecrit_rapport("Pass %u +%u file%s, ",pass,file_nbr-old_file_nbr,(file_nbr-old_file_nbr<=1?"":"s"));
    if(file_nbr>1)
    {
      ecrit_rapport("%u pictures or videos found\n",file_nbr);
    }
    else
    {
      ecrit_rapport("%u picture or video found\n",file_nbr);
    }
  }
  {
    unsigned long int nbr_headers=0;
    uint64_t sectors_with_unknown_data=0;
    /* Free memory */
    t_alloc_data *next;
    for(current_free_space=list_file;current_free_space!=NULL;current_free_space=next)
    {
      next=current_free_space->next;
      FREE(current_free_space);
    }
    for(current_free_space=list_free_space;current_free_space!=NULL;current_free_space=next)
    {
      if(current_free_space->extension!=NULL)
	nbr_headers++;
      sectors_with_unknown_data+=(current_free_space->end-current_free_space->start+disk_car->sector_size-1)/disk_car->sector_size;
      if(debug>0)
      {
	ecrit_rapport("%lu-%lu: %s\n",(long unsigned)(current_free_space->start/disk_car->sector_size),
	    (long unsigned)(current_free_space->end/disk_car->sector_size), current_free_space->extension);
#ifdef OLD
	if(debug>0)
	{
	  for(offset=current_free_space->start;offset<current_free_space->end; offset+=blocksize)
	  {
	    unsigned char buffer[blocksize];
	    if(disk_car->read(disk_car,blocksize, &buffer, offset)!=0)
	    { /* Read error */
	      memset(buffer,0,blocksize);
	    }
	    dump_log(buffer,blocksize);
	  }
	}
#endif
      }
      next=current_free_space->next;
      FREE(current_free_space);
    }
    ecrit_rapport("%llu sectors contains unknown data, %lu invalid files found and rejected.\n",
	(long long unsigned)sectors_with_unknown_data, (long unsigned)nbr_headers);
  }
  wmove(stdscr,9,0);
  wclrtoeol(stdscr);
  wdoprintf(stdscr,"%u pictures or videos found\n",file_nbr);
  wmove(stdscr,10,0);
  wclrtoeol(stdscr);
  if(ind_stop==1)
    ecrit_rapport("PhotoRec has been stopped, ");
  aff_buffer(BUFFER_DISPLAY,"Q",stdscr);
  return 0;
}

static int do_curses_photorec(int debug, const char *recup_dir, const t_list_disk *list_disk, const t_arch_fnct **arch)
{
  int command;
  int done=0;
  int allow_partial_last_cylinder=0;
  int halt_on_errors=1;
  int paranoid=1;
  int menu=0;
  int keep_corrupted_file=0;
  const t_list_disk *element_disk;
  const t_list_disk *current_disk=list_disk;
  static struct MenuItem menuMain[]=
  {
	{'S',"Search","Search pictures"},
	{'O',"Options","Modify options"},
	{'Q',"Quit","Quit program"},
	{0,NULL,NULL}
  };
  if(list_disk==NULL)
    return 0;
/* use_env(TRUE); */
#ifdef HAVE_SETENV
#ifdef DJGPP
#elif defined(BSD)
    setenv("TERM","cons25",0);
#elif defined(LINUX)
    setenv("TERM","linux",0);
#elif defined(__CYGWIN__)
    setenv("TERM","cygwin",0);
#endif
#endif
#ifdef DJGPP
  initscr();
#else
  if(newterm(NULL,stdout,stdin)==NULL)
  {
#ifdef HAVE_SETENV
    setenv("TERMINFO",".",1);
#endif
    if(newterm(NULL,stdout,stdin)==NULL)
    {
      ecrit_rapport("Terminfo file is missing\n");
      printf("Terminfo file is missing\n");
      return 1;
    }
  }
#endif
  if(LINES<25)
  {
    ecrit_rapport("Terminal has only %u lines\n",LINES);
    endwin();
    return 1;
  }
  noecho();
#ifndef DJGPP
  nonl(); /*don't use for Dos version but enter will work with it... dilema */
#endif
  /*  intrflush(stdscr, FALSE); */
  cbreak();
  /* Should solve a problem with users who redefined the colors */
  if(has_colors())
  {
    start_color();
  }
  curs_set(0);
#ifdef HAVE_MKDIR
#ifdef __MINGW32__
  if(mkdir(recup_dir)!=0 && errno==EEXIST)
#else
  if(mkdir(recup_dir, 0700)!=0 && errno==EEXIST)
#endif
  {
    display_message("WARNING: recovered pictures will overwrite the previous ones.");
  }
#else
  {
    struct stat stat_info;
    if(stat(recup_dir, &stat_info)<0)
    {
      display_message("WARNING: No directory to recover the pictures");
      done=1;
    }
    else
    {
      if(S_ISDIR(stat_info.st_mode))
      {
	display_message("WARNING: No directory to recover the pictures");
	done=1;
      }
    }
  }
#endif
  /* ncurses interface */
  while(done==0)
  {
    aff_copy(stdscr);
    wmove(stdscr,5,0);
    for(element_disk=list_disk;element_disk!=NULL;element_disk=element_disk->next)
    {
      if(element_disk!=current_disk)
	wdoprintf(stdscr,"%s\n",element_disk->disk->description(element_disk->disk));
      else
      {
	wattrset(stdscr,A_STANDOUT);
	wdoprintf(stdscr,"%s\n",element_disk->disk->description(element_disk->disk));
	wattroff(stdscr,A_STANDOUT);
      }
    }
    current_disk->disk->halt_on_errors=halt_on_errors;
    command = wmenuSelect(stdscr,INTER_MAIN_Y, INTER_MAIN_X, menuMain, 8,
	  "SOQ", MENU_HORIZ | MENU_BUTTON | MENU_ACCEPT_OTHERS, menu);
    switch(command)
    {
      case 's':
      case 'S':
	interface_select(current_disk->disk, debug, paranoid,recup_dir,keep_corrupted_file);
	menu=2;
	break;
      case 'q':
      case 'Q':
	done = 1;
	break;
      case 'o':
      case 'O':
	{
	  int old_allow_partial_last_cylinder=allow_partial_last_cylinder;
	  interface_options_photorec(&paranoid,&allow_partial_last_cylinder,&halt_on_errors,arch,&keep_corrupted_file);
	  for(element_disk=list_disk;element_disk!=NULL;element_disk=element_disk->next)
	  {
	    element_disk->disk->arch=*arch;
	  }
	  if(old_allow_partial_last_cylinder!=allow_partial_last_cylinder)
	    hd_parse_bis(list_disk,allow_partial_last_cylinder,*arch);
	  menu=1;
	}
	break;
      case MENU_UP:
	if(current_disk->prev!=NULL)
	  current_disk=current_disk->prev;
	break;
      case MENU_DOWN:
	if(current_disk->next!=NULL)
	  current_disk=current_disk->next;
	break;
    }
  }
  wclear(stdscr);
  wrefresh(stdscr);
  nl();
  endwin();
  ecrit_rapport("\n");
  return 0;
}

int main( int argc, char **argv )
{
  int i;
  int help=0, create_log=0, debug=0;
  const char *recup_dir="recup_dir";
  t_list_disk *list_disk=NULL;
  t_list_disk *element_disk;
#ifdef SOLARIS
  const t_arch_fnct *arch=&arch_sun;
#else
  const t_arch_fnct *arch=&arch_i386;
#endif
#ifdef HAVE_SIGACTION
  struct sigaction action, old_action;
#endif
#ifdef TESTING
  srand(1);
#endif
#ifdef HAVE_SIGACTION
  /* set up the signal handler for SIGHUP */
  action.sa_handler  = sighup_hdlr;
  action.sa_flags = 0;
  if(sigaction(SIGHUP, &action, &old_action)==-1)
  {
    printf("Error on SIGACTION call\n");
    return -1;
  }
#endif
#ifdef __CYGWIN__
  /* Average Windows user doesn't know how to specify a parameter */
  if(argc==1)
  {
    create_log=1;
    debug=1;
  }
#endif
  printf("PhotoRec %s, Data Recovery Utility, %s\nChristophe GRENIER <grenier@cgsecurity.org>\nhttp://www.cgsecurity.org\n",VERSION,TESTDISKDATE);
  for(i=1;i<argc;i++)
  {
    if((strcmp(argv[i],"/log")==0) ||(strcmp(argv[i],"-log")==0))
      create_log=1;
    else if((strcmp(argv[i],"/debug")==0) || (strcmp(argv[i],"-debug")==0))
    {
      debug++;
      create_log=1;
    }
    else if(((strcmp(argv[i],"/d")==0)||(strcmp(argv[i],"-d")==0)) &&(i+1<argc))
    {
      recup_dir=argv[i+1];
      i++;
    }
    else if((strcmp(argv[i],"/help")==0) || (strcmp(argv[i],"-help")==0) || (strcmp(argv[i],"--help")==0) ||
      (strcmp(argv[i],"/h")==0) || (strcmp(argv[i],"-h")==0))
      help=1;
    else
    {
      list_disk=insert_new_disk(list_disk,file_test_availability(argv[i],debug,arch));
      if(list_disk==NULL)
      {
	printf("\nUnable to open file or device %s\n",argv[i]);
	help=1;
      }
    }
  }
  if(help!=0)
  {
    printf("\nUsage: photorec [/log] [/debug] [/d recup_dir] [file or device]"\
	    "\n" \
	    "\n/log          : create a photorec.log file" \
	    "\n/debug        : add debug information" \
	    "\n" \
	    "\nPhotoRec searches JPEG and MOV files, it stores them in recup_dir directory." \
	    "\n" \
	    "\nIf you have problems with PhotoRec or bug reports, please contacte me.\n");
    return 0;
  }
  if(create_log!=0)
  {
/*    const char *ext2fs_version=NULL; */
    f_rapport=init_log("photorec.log",argc,argv);
  ecrit_rapport("PhotoRec %s, Data Recovery Utility, %s\nChristophe GRENIER <grenier@cgsecurity.org>\nhttp://www.cgsecurity.org\n",VERSION,TESTDISKDATE);
#ifdef DJGPP
    ecrit_rapport("Dos version");
#elif defined(BSD)
    ecrit_rapport("BSD version");
#elif defined(LINUX)
    ecrit_rapport("Linux version");
#elif defined(__CYGWIN__)
    ecrit_rapport("Windows version");
#else
    ecrit_rapport("Undefined OS");
#endif
    ecrit_rapport("\n");
  }
  printf("Please wait...\n");
  {
    const char *locale;
    locale = setlocale (LC_ALL, "");
    if (locale==NULL) {
      locale = setlocale (LC_ALL, NULL);
      ecrit_rapport("Failed to set locale, using default '%s'.\n", locale);
    } else {
      ecrit_rapport("Using locale '%s'.\n", locale);
    }
  }
  aff_buffer(BUFFER_RESET,"Q");
  list_disk=hd_parse(list_disk,debug,arch);
  if(list_disk==NULL)
  {
    printf("No harddisk found\n");
#ifdef __CYGWIN__
    printf("You need to be administrator to use PhotoRec\n");
    printf("Under Win9x, use the DOS version instead\n");
    ecrit_rapport("You need to be administrator to use PhotoRec\n");
#else
#ifndef DJGPP
#ifdef HAVE_GETEUID
    if(geteuid()!=0)
    {
      printf("You need to be root to use PhotoRec\n");
      ecrit_rapport("You need to be root to use PhotoRec\n");
    }
#endif
#endif
#endif
  }
#ifdef DJGPP
  for(element_disk=list_disk;element_disk!=NULL;element_disk=element_disk->next)
  {
    printf("%s\n",element_disk->disk->description(element_disk->disk));
  }
#endif
  hd_parse_bis(list_disk,0,arch);
  /* save disk parameters to rapport */
  ecrit_rapport("Hard disk list\n");
  for(element_disk=list_disk;element_disk!=NULL;element_disk=element_disk->next)
  {
    printf("%s, sector size=%u\n",element_disk->disk->description(element_disk->disk),element_disk->disk->sector_size);
    ecrit_rapport("%s, sector size=%u\n",element_disk->disk->description(element_disk->disk),element_disk->disk->sector_size);
  }
  printf("\n");
  ecrit_rapport("\n");

  {
    if(do_curses_photorec(debug,recup_dir,list_disk,&arch))
    {
      printf("PhotoRec need 25 lines to work.\nPlease enlarge the terminal and restart PhotoRec.\n");
    }
  }
  for(element_disk=list_disk;element_disk!=NULL;)
  {
    t_list_disk *element_disk_next=element_disk->next;
    if(element_disk->disk->clean!=NULL)
      element_disk->disk->clean(element_disk->disk);
    FREE(element_disk->disk);
    FREE(element_disk);
    element_disk=element_disk_next;
  }
  if(f_rapport!=NULL)
  {
    ecrit_rapport("PhotoRec exited normally.\n");
    if(fclose(f_rapport))
    {
      f_status=1;
    }
  }
  if(f_status!=0)
  {
    printf("PhotoRec: Log file corrupted!\n");
  }
  else
  {
    printf("PhotoRec exited normally.\n");
  }
  return 0;
}

static void interface_options_photorec(int *paranoid, int *allow_partial_last_cylinder,int *halt_on_errors, const t_arch_fnct **arch, int *keep_corrupted_file)
{
  int done = FALSE;
  int menu = 5;
  while (done==FALSE)
  {
    int car;
    struct MenuItem menuOptions[]=
    {
      { 'P', NULL, "Check JPG files" },
      { 'A',NULL,"" },
      { 'H',NULL,"Halt on Disk I/O errors" },
      { 'T',NULL,"Partition table type"},
      { 'K',NULL,"Keep corrupted files"},
      { 'Q',"Quit","Return to main menu"},
      { 0, NULL, NULL }
    };
    menuOptions[0].name=*paranoid?"Paranoid : Yes":"Paranoid : No";
    menuOptions[1].name=*allow_partial_last_cylinder?"Allow partial last cylinder : Yes":"Allow partial last cylinder : No";
    menuOptions[2].name=*halt_on_errors?"Halt on errors : Yes":"Halt on errors : No";
    menuOptions[3].name=(*arch==&arch_i386?"Intel":(*arch==&arch_sun?"Sun":"None"));
    menuOptions[4].name=*keep_corrupted_file?"Keep corrupted files : Yes":"Keep corrupted files : No";
    /* Jpg
       Mov
       Mpg
       Minolta MRW
       Canon CRW
       Signa/Foveon X3F
       Fuji RAF
       Rollei RDC
       MP3

    */
    car=toupper( wmenuSelect(stdscr,INTER_OPTION_Y, INTER_OPTION_X, menuOptions, 0, "PAHTKQ", MENU_VERT, menu));
    switch(car)
    {
      case 'P':
	*paranoid=!*paranoid;
	menu=0;
	break;
      case 'A':
	*allow_partial_last_cylinder=!*allow_partial_last_cylinder;
	menu=1;
	break;
      case 'H':
	*halt_on_errors=!*halt_on_errors;
	menu=2;
	break;
      case 'T':
       if(*arch==&arch_i386)
	 *arch=&arch_sun;
       else if(*arch==&arch_sun)
	 *arch=&arch_none;
       else
	 *arch=&arch_i386;
	menu=3;
	break;
      case 'K':
	*keep_corrupted_file=!*keep_corrupted_file;
	menu=4;
	break;
      case key_ESC:
      case 'Q':
	done = TRUE;
	break;
    }
  }
  /* write new options to log file */
  ecrit_rapport("New options :\n Paranoid : %s\n",
      *paranoid?"Yes":"No");
  ecrit_rapport("\n Allow partial last cylinder : %s\n Halt on errors : %s\n Partition table type : %s\nKeep corrupted files : %s\n",
      *allow_partial_last_cylinder?"Yes":"No",
      *halt_on_errors?"Yes":"No",
      (*arch==&arch_i386?"Intel":(*arch==&arch_sun?"Sun":"None")),
      *keep_corrupted_file?"Yes":"No");
}

static void interface_select(t_param_disk *disk_car, const int debug,const int paranoid, const char *recup_dir, const int keep_corrupted_file)
{
  t_partition *fake_partition;
  t_list_part *list_part;
  t_list_part *element;
  t_list_part *current_element;
  int quit;
  int offset=0;
  int current_element_num=0;
  int rewrite=1;
  int menu=0;
  ecrit_rapport("\nInterface Select\n");
  list_part=disk_car->arch->read_part(disk_car,debug,0);
  fake_partition=partition_new();
  fake_partition->part_offset=0;
  fake_partition->part_size=disk_car->disk_size;
  fake_partition->part_type=0;
  strncpy(fake_partition->name,"Whole disk",sizeof(fake_partition->name)-1);
  list_part=insert_new_partition(list_part,fake_partition);
  current_element=list_part;
  for(element=list_part;element!=NULL;element=element->next)
  {
    aff_part_rapport(disk_car,element->part);
  }
  do
  {
    static struct MenuItem menuAdv[]=
    {
      {'s',"Search","Start photo recovery"},
      {'q',"Quit","Return to main menu"},
      {0,NULL,NULL}
    };
    int i;
    int command;
    if(rewrite!=0)
    {
      aff_copy(stdscr);
      wmove(stdscr,4,0);
      wdoprintf(stdscr,"%s",disk_car->description(disk_car));
      mvwaddstr(stdscr,6,0,msg_PART_HEADER_LONG);
      rewrite=0;
    }
    for(i=0,element=list_part;(element!=NULL) && (i<offset);element=element->next,i++);
    for(i=offset;(element!=NULL) && ((i-offset)<INTER_SELECT);i++,element=element->next)
    {
      wmove(stdscr,5+2+i-offset,0);
      wclrtoeol(stdscr);	/* before addstr for BSD compatibility */
      if(element==current_element)
      {
	wattrset(stdscr,A_STANDOUT);
	aff_part(stdscr,AFF_PART_ORDER,disk_car,element->part);
	wattroff(stdscr,A_STANDOUT);
      } else
      {
	aff_part(stdscr,AFF_PART_ORDER,disk_car,element->part);
      }
    }
    quit=0;
    command = wmenuSelect(stdscr,INTER_SELECT_Y, INTER_SELECT_X, menuAdv, 8,
	"sq", MENU_HORIZ | MENU_BUTTON | MENU_ACCEPT_OTHERS, menu);
    switch(command)
    {
      case MENU_UP:
	if(current_element!=NULL)
	{
	  if(current_element->prev!=NULL)
	  {
	    current_element=current_element->prev;
	    current_element_num--;
	  }
	  if(current_element_num<offset)
	    offset--;
	}
	break;
      case MENU_DOWN:
	if(current_element!=NULL)
	{
	  if(current_element->next!=NULL)
	  {
	    current_element=current_element->next;
	    current_element_num++;
	  }
	  if(current_element_num>=offset+INTER_SELECT)
	    offset++;
	}
	break;
      case 'q':
      case 'Q':
	quit=1;
	break;
      case 's':
      case 'S':
	if(current_element!=NULL)
	{
	  photorec(disk_car, current_element->part, debug, paranoid, recup_dir,keep_corrupted_file);
	  rewrite=1;
	  menu=1;
	}
	break;
    }
  } while(quit==0);
  delete_list_part(list_part);
}

void aff_copy(WINDOW *window)
{
  wclear(window);
  keypad(window, TRUE); /* Need it to get arrow key */
  wmove(window,0,0);
  wdoprintf(window, "PhotoRec %s, Data Recovery Utility, %s\n",VERSION,TESTDISKDATE);
  wmove(window,1,0);
  wdoprintf(window, "Christophe GRENIER <grenier@cgsecurity.org>");
  wmove(window,2,0);
  wdoprintf(window, "http://www.cgsecurity.org");
}

#ifdef HAVE_LIBJPEG
struct my_error_mgr {
  struct jpeg_error_mgr pub;	/* "public" fields */

  jmp_buf setjmp_buffer;	/* for return to caller */
};

typedef struct my_error_mgr * my_error_ptr;
static void my_output_message (j_common_ptr cinfo);
static void my_error_exit (j_common_ptr cinfo);

static void my_output_message (j_common_ptr cinfo)
{
#ifdef DEBUG
  char buffermsg[JMSG_LENGTH_MAX];
  /* Create the message */
  (*cinfo->err->format_message) (cinfo, buffermsg);
  ecrit_rapport("test_jpeg error %s\n",buffermsg);
#endif
}

static void my_error_exit (j_common_ptr cinfo)
{
  my_error_ptr myerr = (my_error_ptr) cinfo->err;
  (*cinfo->err->output_message) (cinfo);
  /* Return control to the setjmp point */
  longjmp(myerr->setjmp_buffer, 1);
}

static int64_t test_jpeg(FILE *infile)
{
  struct jpeg_decompress_struct cinfo;
  JSAMPARRAY buffer;		/* Output row buffer */
  int row_stride;		/* physical row width in output buffer */
  uint64_t jpeg_size=0;
  struct my_error_mgr jerr;
  fseek(infile,SEEK_SET,0);
  cinfo.err = jpeg_std_error(&jerr.pub);
  jerr.pub.output_message = my_output_message;
  jerr.pub.error_exit = my_error_exit;
  /* Establish the setjmp return context for my_error_exit to use. */
  if (setjmp(jerr.setjmp_buffer)) {
    /* If we get here, the JPEG code has signaled an error.
     * We need to clean up the JPEG object and return.
     */
/*    ecrit_rapport("JPG error at %llu\n",(long long unsigned)ftell(infile)); */
    jpeg_destroy_decompress(&cinfo);
    /* Not accurate */
    return -ftell(infile);
  }
  jpeg_create_decompress(&cinfo);
  cinfo.two_pass_quantize = FALSE;
  cinfo.dither_mode = JDITHER_NONE;
  cinfo.desired_number_of_colors = 2;
  cinfo.dct_method = JDCT_FASTEST;
  cinfo.do_fancy_upsampling = FALSE;
  cinfo.raw_data_out = TRUE;

  jpeg_stdio_src(&cinfo, infile);
  (void) jpeg_read_header(&cinfo, TRUE);
  (void) jpeg_start_decompress(&cinfo);
  row_stride = cinfo.output_width * cinfo.output_components;
  buffer = (*cinfo.mem->alloc_sarray)
		((j_common_ptr) &cinfo, JPOOL_IMAGE, row_stride, 1);
  while (cinfo.output_scanline < cinfo.output_height) {
    (void) jpeg_read_scanlines(&cinfo, buffer, 1);
  }
  (void) jpeg_finish_decompress(&cinfo);
  jpeg_destroy_decompress(&cinfo);
  jpeg_size=ftell(infile);
  if(jerr.pub.num_warnings>0)
  {
    /* Not accurate */
    jpeg_size=-jpeg_size;
/*    ecrit_rapport("JPG error at %llu\n",(long long unsigned)ftell(infile)); */
  }
  return jpeg_size;
}
#else
static int64_t test_jpeg(FILE *infile)
{
  return ftell(infile);
}
#endif

static unsigned int find_blocksize(t_alloc_data *list_file, const unsigned int default_blocksize)
{
  t_alloc_data *current_file;
  int blocksize_ok=0;
  unsigned int blocksize;
  uint64_t offset;
  if(list_file==NULL)
    return default_blocksize;
  for(blocksize=64*512;blocksize>default_blocksize && blocksize_ok==0;blocksize=blocksize>>1)
  {
    blocksize_ok=1;
    offset=list_file->start%blocksize;
    for(current_file=list_file;current_file!=NULL && current_file->next!=NULL && blocksize_ok!=0;current_file=current_file->next)
    {
      if(current_file->start%blocksize!=offset)
	blocksize_ok=0;
    }
  }
  blocksize=blocksize<<1;
  return blocksize;
}

static void update_blocksize(unsigned int blocksize, t_alloc_data *list_free_space)
{
  t_alloc_data *current_free_space;
  uint64_t offset;
  if(list_free_space==NULL)
    return ;
  offset=list_free_space->end%blocksize;
  for(current_free_space=list_free_space;current_free_space!=NULL;current_free_space=current_free_space->next)
  {
    current_free_space->start=(current_free_space->start-(offset+1)%blocksize+blocksize-1)/blocksize*blocksize+(offset+1)%blocksize;
/*    current_free_space->end=(current_free_space->end-(offset)%blocksize+blocksize-1)/blocksize*blocksize+(offset)%blocksize; */
  }
}


